''' Connect, access, + manipulate external tree data from a remote SQL server or
from a sqlite file. '''

# Date:   Nov 7 2013
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

import MySQLdb as mysql, sqlite3 as sqllite
from landscape import landscape, choice
from tree import tree
from abc import ABCMeta as abstractclass, abstractmethod

class DatabaseLandscape(landscape):
    
    ''' Abstract the landscape to one comprising a landscape. '''
    
    __metaclass__ = abstractclass

    def _acquireNumberOfTaxa(self):
        
        topo = self.getTree(0).toTopology()
        self.leaves = len(topo.getLeaves())
        
    @abstractmethod
    def _fetchTreeFromDatabase(self,i): pass
    
    @abstractmethod
    def _fetchRearrangementsFromDatabase(self,i): pass

    def getNode(self,i):
        
        self._fetchTreeFromDatabase(self,i)
        return super(DatabaseLandscape,self).getNode(self,i)

class SQLExhaustiveLandscape(DatabaseLandscape):
    
    def __init__(self,dbobj,aliname):
        
        # Set up fields.
        super(SQLLandscape,self).__init__(None,None,False)
        self.database = dbobj
        self.aliname  = aliname
        self.queryID  = lambda d: dbobj.filterRecords(
            'newick','ID=%d' % (d)) # Alias lookup function.
        self.queryScr = lambda d: dbobj.filterRecords(
            aliname,'TID=%d' % (d))
        self.tables   = dbobj.getTables()
        if not aliname in self.tables:
            raise IOError('Could not find alignment table <%s>.' % (aliname))
        self._fetchTreeFromDatabase(0) # Get root tree.
        self._acquireNumberOfTaxa()
    
    def _fetchTreeFromDatabase(self,i):
        
        # Overrides lookup in graph object first. Looks in database.
        if i in self.graph.node: return
        entries = self.queryID(i)
        if len(entries) == 0: return   
        entry = entries[0]
        
        # Create tree object.
        t = tree(entry[1])
        t.setName(i)
        self.addTree(t) 
        
        # Get relevant score(s).
        entry = self.queryScr(i)[0]
        ml,p  = entry[2:4]
        t.score = (float(ml),float(p))
    
    def _fetchRearrangementsFromDatabase(self,i):
        moves = self.database.filterRecords('operations','TID=%d'%(i))
        return [(int(x),int(y),_,int(z)) for x,y,_,z in moves]
        
    def _applyRearrangement(self,move):
        e,v,o = move[1],move[-1],move[2]
        self.graph.add_edge(e,v)
        self.getTree(v).origin = o
        
    def exploreRandomTree(self,i):
        node  = self.getNode(i)
        if (node['explored']): return None
        moves = self._fetchRearrangementsFromDatabase(i)
        while (len(moves) > 0):
            move = choice(moves)
            moves.remove(move)
            j = move[-1]
            if j in self.graph.node and self.graph.has_edge(i,j):
                continue
            self._fetchTreeFromDatabase(j)
            self._applyRearrangement(move)
            self.getEdge(i,j)['weight'] = self.defaultWeight
            return j
        node['explored'] = True
    
    def getDatabaseNode(self,i):
        self._fetchTreeFromDatabase(i)
        return super(SQLLandscape,self).getNode(i)
    
    def exploreTree(self,i):
        node  = self.getNode(i)
        if (node['explored']): return None
        moves = self._fetchRearrangementsFromDatabase(i)
        while (len(moves) > 0):
            move = choice(moves)
            moves.remove(move)
            j = move[-1]
            if j in self.graph.node and self.graph.has_edge(i,j):
                continue
            self._fetchTreeFromDatabase(j)
            self._applyRearrangement(move)
            self.getEdge(i,j)['weight'] = self.defaultWeight
        node['explored'] = True     
        return True

class SQLiteLandscape(landscape):
    
    ''' Allow random access of all landscape data from an sqlite file found on
    the hard disk. '''
    
    def __init__(self,dbobj):
        
        pass
        

class database(object):

    ''' Allow interfacing with a SQL/sqlite database. '''
    
    __metaclass__ = abstractclass
    cursor = None # Cursor for the database object.
    
    @abstractmethod
    def getTables(self): pass
    
    @abstractmethod
    def getColumns(self,table): pass  
    
    def isEmpty(self):
        ''' Determine if the database is empty. '''
        return (len(self.getTables()) == 0)
    def getHeaders(self,table):
        ''' Get only header names for a given table's columns. '''
        return [x[0] for x in self.getColumns(table)]
    def getRecordsColumn(self,table,col):
        ''' Get all data for a single colmun from records for a table. '''
        self.query("""SELECT %s FROM %s""" % (col,table))
        return self.cursor.fetchall()
    def getRecords(self,table):
        ''' Get all records from a given table in the database. '''
        self.query("""SELECT * FROM %s""" % (table))
        return self.cursor.fetchall()
    def iterRecords(self,table):
        ''' Get a record, one at a time, from a table in the database. '''
        self.query("""SELECT * FROM %s""" % (table))
        nextitem = self.cursor.fetchone()
        while nextitem != None:
            yield nextitem
            nextitem = self.cursor.fetchone()
    def filterRecords(self,table,condn):
        ''' Get all records from a given table following a condition. '''
        self.query("""SELECT * FROM %s WHERE %s""" % (table,condn))
        return self.cursor.fetchall()
    def getRecordsAsDict(self,table):
        ''' Acquires records using getRecords() and then leverages
        access using a dictionary data structure. '''
        d = {}
        items = self.getRecords(table)
        if len(items) == 0: return {}
        headers = self.getHeaders(table)
        if len(items[-1]) != len(headers): return None
        for header in headers: d[header] = []
        for item in items:
            for i in xrange(len(headers)):
                header = headers[i]
                value  = item[i]
                d[header].append(value)
        return d
    def newTable(self,tablename,*args):
        self.query("""CREATE TABLE %s (%s)""" % (
            tablename,', '.join(['%s %s' % (x,y) for x,y in args])))
    def insertRecords(self,tablename,items):
        self.querymany('''INSERT INTO %s VALUES (%s)''' % (
            tablename,', '.join(['?' for x in items[0]])),items)
    def insertRecord(self,tablename,record):
        self.insertRecords(tablename,[record])

    @abstractmethod
    def query(self,q): pass
    
    @abstractmethod
    def querymany(self,q,i): pass

    @abstractmethod
    def close(self): pass

class SQLDatabase(database):
    
    ''' Database object to allow reading from a MySQL database. '''
    
    def __init__(self,host,user,pw,db):
        self.hostname = host
        self.username = user
        self.password = pw
        self.database = db
        self.connect()
    
    def connect(self):
        self.socket   = mysql.connect(host=self.hostname,
                                      user=self.username,
                                      passwd=self.password,
                                      db=self.database)
        self.cursor   = self.socket.cursor()        
        
    def getTables(self):
        self.query("""SHOW TABLES""")
        return [x[0] for x in self.cursor.fetchall()]        
        
    def getColumns(self,table):
        ''' Return column information for a given table. '''
        self.query("""SHOW COLUMNS FROM %s""" % (table))
        return self.cursor.fetchall()          
        
    def query(self,q):
        # Execute a MySQL query.
        try:
            self.cursor.execute(q)
        except (AttributeError,mysql.OperationalError):
            # In case of server timeout, reconnect.
            self.connect()
            self.query(q)        

    def querymany(self,q,i):
        try:
            self.cursor.executemany(q,i)
        except (AttributeError,mysql.OperationalError):
            self.connect()
            self.querymany(q,i)

    def close(self):
        # Close the connection.
        self.socket.close()
        

class SQLiteDatabase(database):
    
    def __init__(self,filepath):
        self.filepath = filepath
        self.socket   = sqllite.connect(filepath)
        self.cursor   = self.socket.cursor()

    def getColumns(self,table):
        ''' Return column information for a given table. '''
        self.query("""PRAGMA table_info(%s)""" % (table))
        return self.cursor.fetchall()

    def getTables(self):
        self.query("""SELECT name FROM sqlite_master WHERE type='table';""")
        return [x[0] for x in self.cursor.fetchall()]   

    def query(self,q):
        # Execute an sqlite query.
        self.cursor.execute(q)

    def querymany(self,q,i):
        self.cursor.executemany(q,i)

    def close(self):
        # Close the connection.
        self.socket.commit()
        self.socket.close()    

        



''' Toolkit for performance of Parsimonious Criterion (Parsimony) methods of
optimization of a phylogenetic topology with a particular set of data. '''

# Date:   Mar 3 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

from base import treeStructure
postorder = treeStructure.postOrderTraversal

class profile_set:
    
    ''' Hold a set of site_profile profiles for an 
    entire alignment. '''

    def __init__(self,alignment):
        
        self.alignment = alignment
        self.numSites  = alignment.getSize()
        self.taxa      = {}
        self.profiles  = []
        self.weights   = []
        self.sites     = []
        self._constructSet()
        self._buildTaxaDict()
    
    def _constructSet(self):
        
        ''' PRIVATE: Build all site_profiles. '''
        
        num = self.numSites
        for site in xrange(num):
            pro = site_profile(self.alignment,site)
            self.sites.append(pro)
            if pro in self.profiles:
                self.weights[self.profiles.index(pro)] += 1
            else:
                self.profiles.append(pro)
                self.weights.append(1)
                
    def _buildTaxaDict(self):
        
        ''' PRIVATE: Build taxa dictionary. '''
        
        tax = self.alignment.getTaxa()
        for t in xrange(len(tax)): self.taxa[tax[t]] = t
                
    def __len__(self):
        
        return len(self.profiles)
    
    def weight(self,val):
        
        return self.weights[val]
    
    def get(self,val):
        
        return self.profiles[val]
    
    def getForTaxa(self,val,tax):
        
        return self.profiles[val].vector[self.taxa[tax]]

class site_profile:
    
    ''' Consolidate the single-column alignment at
    a region into a set of components on the basis
    of similarity alone. '''
    
    def __init__(self,alignment,site):
        
        self.alignment = alignment
        self.site      = site
        self.alphabet  = None
        self.vector    = ''
        self._buildVector()
        
    def __eq__(self,o):
        
        if (o == None): return False
        return (self.vector == o.vector)
    
    def __ne__(self,o):
        
        return not self.__eq__(o)
        
    def __str__(self):
        
        return self.vector
        
    def _buildVector(self):
        
        ''' Build the vector comprising the set of components. '''
        
        aliData = self.alignment.data
        sitesli = aliData.sequenceSlice(self.site)
        compn   = 0
        comps   = {}
        for it in sitesli:
            if not it in comps:
                comps[it] = compn
                compn += 1
            self.vector += str(comps[it])
        self.alphabet = comps

def fitch_cost(topology,profiles):
    
    ''' Calculate the cost using Fitch algorithm on 
    profile set and alignment. Deprecated: Python implementation 
    of the Fitch algorithm; see fitch C++ module for 
    a C++ implementation that is roughly four times faster. '''
    
    # Calculate parsimony score.
    total = 0
    porde = [x for x in postorder(topology.getRoot())]
    for profile in xrange(len(profiles)):
        posdata, local = {}, 0
        for node in porde:
            # If is leaf.
            if (len(node.children) == 0):
                posdata[node] = profiles.getForTaxa(profile,node.label)
            # Not a leaf.
            else:
                a = posdata[node.children[0].child]
                b = posdata[node.children[1].child]
                # Get intersection.
                X = set(a).intersection(set(b))
                # If intersection is nothing...
                if (len(X) == 0):
                    # Cost goes up, take union.
                    local += 1
                    X = set(a).union(set(b))
                posdata[node] = X
        total += local*profiles.weight(profile)
    return total

def fitch(topology,alignment):
        
    ''' Perform the Fitch algorithm on a given
    tree topology and associated alignment. Deprecated: 
    Python implementation of the Fitch algorithm; 
    see fitch C++ module for a C++ implementation 
    that is roughly four times faster. '''
    
    # Turn alignment sets into profiles.
    profiles = profile_set(alignment)
    if (len(profiles) == 0): return None
    return fitch_cost(topology,profiles)


''' Handle input biological sequence alignment files for the purposes of
phylogenetic inference. Will read all types of alignment files by utilizing the 
P4 python phylogenetic library. '''

# Date:   Jan 24 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

# Imports

import os
import p4
import model
import tree
import base
from newick import newickParser
from executable import fasttree
from tempfile import NamedTemporaryFile as NTempFile
from shutil import copyfile

# Set up.

p4.var.warnReadNoFile = False

# Class Definitions


class alignment(object):
    
    ''' Wrap a biological sequence alignment to enable functionality
    necessary for phylogenetic inference. Makes use of temporary
    files; requires to be closed once no longer needed. '''
    
    def __init__(self,inal=None):
        
        ''' Instantiate an object intended to wrap an alignment for the
        purposes of running phylogenetic inference. 
        
        :param inal: An alignment file path (most formats are accepted). 
        
        '''
        
        # Intergrate the p4 phylogenetic library.
        if inal is None:
            self.data = p4.Alignment()
            
        else:
            p4.read(inal) # Read the alignment file/string.
            self.data  = p4.var.alignments[-1]

        # Augment alignment with a discrete state model.
        self.model = model.DiscreteStateModel(self)

        # Keep track of temporary files.
        self.paths = {} 

    # Internals

    def __getitem__(self,i): return self.data.sequences[i]
    def __str__(self):       return '\n'.join(self.toStrList())
    def __len__(self):       return len(self.data)
    
    def __iter__(self):
        
        for i in xrange(self.getNumSeqs()): yield self[i]

    def _makeFASTAFile(self):
        
        ''' PRIVATE: Make a temporary FASTA file. '''
        
        if ('fasta' in self.paths):
            return
        fh = NTempFile(delete=False)
        self.paths['fasta'] = fh.name
        self.data.writeFasta(fh.name)
        fh.close()

    def close(self):
        
        ''' Delete all temporary files and clear data. '''
        
        self.data  = None
        self.model = None
        if len(self.paths) > 0:
            for f in self.paths.values(): os.unlink(f)
            self.paths = {}

    # Accessors

    def toStrList(self):

        ''' Get all sequences as a list of strings. '''

        return [self.getSequence(i) for i in xrange(self.getNumSeqs())]

    def getStateModel(self):

        return self.model

    def getSize(self):

        ''' Return the size of the alignment, or how many characters
        there are in each respective item in the alignment. '''

        return self.__len__()

    def getNumSeqs(self):

        ''' Return the number of sequences that are present in the 
        sequence alignment. '''

        return len(self.data.sequences)

    def getDim(self):

        ''' Return the dimensionality of the sequence alignment (how 
        many different types of characters). '''

        return self.data.dim

    def getSequence(self,i):

        ''' Acquire the ith sequence. '''

        return self[i].sequence

    def getFASTA(self):

        ''' Get (and create if not already) a path to a temporary 
        FASTA file. This will be deleted upon closure of the alignment
        instance. '''

        self._makeFASTAFile()
        return self.paths['fasta']

    def getApproxMLNewick(self):

        ''' Get a tree in newick format via use of FastTree that serves as
        an approximation of the maximum likelihood tree for this data. '''

        ft = fasttree(
            self.getFASTA(), isProtein=(self.data.dataType=='protein'))
        return ft.run()

    def getApproxMLTree(self):

        ''' Get a tree object for an approximation of the maximum likelihood
        tree for this data using FastTree. '''

        return tree.tree(self.getApproxMLNewick(),check=True)

    def getFastTreeNewick(self):

        ''' Alias for the "getApproxMLNewick()" function. '''

        return self.getApproxMLNewick()

    def getTaxa(self):

        ''' Return taxa names. '''

        return self.data.taxNames

    # Interface Revealing

    def getAlignment(self):

        ''' Acquire the alignment data structure (P4 module). '''

        return self.data

    def bootstrap(self):

        ''' Perform bootstrapping on the alignment data. '''

        self.data = self.data.bootstrap()    


class phylipFriendlyAlignment(alignment):

    ''' An alignment object that renames all comprising taxa in order
    to be able to be written as a strict Phylip file. '''

    def __init__(self,inal=None):

        # Call superclass constructor.
        super(phylipFriendlyAlignment,self).__init__(inal)

        # Keep track of all taxa names.
        self.namedict = {}

        # Acquire a temporary file handle for Nexus with proper names.
        self._makeProperNexusFile()

        # Reassign all taxa names to shorter ones (for Phylip).
        self._reassignNames()  

    def _makePhylipFile(self):

        ''' PRIVATE: Make a temporary Phylip file. '''

        if ('phylip' in self.paths): return
        fh = NTempFile(delete=False)
        self.paths['phylip'] = fh.name
        self.data.writePhylip(
            fh.name,whitespaceSeparatesNames=False)
        fh.close()
        
    def _makeProperNexusFile(self):

        ''' PRIVATE: Make a temporary Nexus file. '''

        if ('nexus' in self.paths): return
        fh = NTempFile(delete=False)
        self.paths['nexus'] = fh.name
        self.data.writeNexus(
            fName=fh.name)    
        fh.close()

    def _reassignNames(self):

        ''' PRIVATE: Reassign all sequence names to shorter 
        Phylip-friendly integer-based names. '''

        d  = self.namedict
        for item in self.data.sequences:
            _         = item.name
            ind       = self.data.sequences.index(item)
            name      = 'T' + str(ind)
            d[name]   = _
            item.name = name
            self.data.taxNames[ind] = name

    def getPhylip(self):

        ''' Get a path to a temporary Phylip file. This will 
        be deleted upon closure of the alignment instance. '''

        self._makePhylipFile()
        return self.paths['phylip']

    def writeProperNexus(self,wri):

        ''' Write a Nexus file with proper names. '''

        copyfile(self.paths['nexus'],wri)
        return True

    def reassignFromReinterpretedNewick(self,tr):

        ''' Replace all proper names with reassigned names in Newick tree. '''

        p = newickParser(tr).parse()
        nodes = base.treeStructure(p).getAllNodes()
        for node in nodes:
            l = node.label
            if l in self.namedict.values():
                for n in self.namedict:
                    if self.namedict[n] == l:
                        node.label = n
                        break
        return str(p) + ';'

    def reinterpretNewick(self,tr):

        ''' Replaces all reassigned names to proper names in Newick tree. '''

        p = newickParser(tr).parse()
        nodes = base.treeStructure(p).getAllNodes()
        for node in nodes:
            l = node.label
            if l in self.namedict: node.label = self.namedict[l]
        return str(p) + ';'

    def getProperName(self,n):

        ''' Return the actual name for an integer-based sequence name 
        that was reassigned at initialization. '''

        if n in self.namedict: return self.namedict[n]
        else: return None

    def getTaxa(self):

        ''' Return current taxa names in the alignment. '''

        return [x for x in self.namedict]

    def recreateObject(self):

        ''' Reintializes the object. '''

        return phylipFriendlyAlignment(self.getFASTA())



''' Definitions for generalized containers and objects used by other structures
in this framework. '''

# Date:   Nov 26 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

# Imports

from collections import Sized, Iterable, Container

# Function Definitions

def longest_common_substring(s1,s2):
    
    ''' Simplified, traditional LCS algorithm implementation. '''
    
    m = [[0] * (1 + len(s2)) for i in xrange(1 + len(s1))]
    longest, x_longest = 0, 0
    for x in xrange(1, 1 + len(s1)):
        for y in xrange(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
                    x_longest = x
            else: m[x][y] = 0
    return s1[x_longest - longest: x_longest]

# Tree Elements

class treeStructure(Container):
    
    ''' Defines a base collection of treeNodes and treeBranches in a
    hierarchical tree structure. '''
    
    root = None
    
    def __init__(self,root=None):
        
        if root == None:
            self.root = treeNode()
        else: self.root = root
    
    def __contains__(self,x):
        
        ''' Determines whether a node is found in the tree
        structure. '''
        
        for node in self.getPostOrderTraversal():
            if node == x: return True
        return False
    
    def getRoot(self):
        
        ''' Return the top-level, root, node of 
        the tree. '''
        
        return self.root    
    
    @staticmethod
    def leaves(root):
        if root.isLeaf(): return [root]
        li = list()
        for branch in root.getChildren():
            li.extend(treeStructure.leaves(branch.getChild()))
        return li        
        
    def getAllLeaves(self): return treeStructure.leaves(self.root)
    
    @staticmethod
    def nodes(root):
        if root == None: return []
        li = list()
        li.append(root)
        for branch in root.getChildren():
            li.extend(treeStructure.nodes(branch.getChild()))
        return li    
    
    def getAllNodes(self): return treeStructure.nodes(self.root)
    
    @staticmethod
    def postOrderTraversal(root):
        if (root == None): return []
        li = list()
        for branch in root.getChildren():
            li.extend(treeStructure.postOrderTraversal(branch.child))
        li.append(root)
        return li
        
    def getPostOrderTraversal(self):
        return treeStructure.postOrderTraversal(self.root)

    def _createStringLinesForNode(self,root,levelsDeep=0):
        if (root == None): return '\n'
        elif (len(root.getChildren()) == 0):
            return '  '*levelsDeep + '| ' + root.label
        strOut = ''
        for branch in root.getChildren():
            strOut += self._createStringLinesForNode(branch.child,levelsDeep+1)
        return strOut
    
    def __str__(self):
        
        ''' Returns a string representation of the tree. '''

        return self._createStringLinesForNode(self.root)

class treeNode(object):
    
    ''' A node in a tree. '''

    label    = None
    parent   = None
    children = None
    
    def __init__(self,lbl=None,children=None,parent=None):
        
        self.label  = lbl
        self.parent = parent
        if children != None:
            self.children = children
        else: self.children = list()
            
    def getLabel(self): return self.label
    def getParent(self): return self.parent
    def addChild(self,item):
        self.children.append(item)
    def getChildByIndex(self,i): return self.children[i]
    def getChildren(self): return self.children
    def isLeaf(self): return len(self.children) == 0
    def isInternalNode(self): return (len(self.children) > 0)

class treeBranch(object):
    
    ''' A branch in a tree. '''
    
    label  = ''    
    parent = None
    child  = None
    
    def __init__(self,parent=None,child=None,label=''):
        
        self.parent = parent
        self.child  = child
        self.label  = label
        
    def getLabel(self): return self.label
    def getParent(self): return self.parent
    def getChild(self): return self.child  

# Tries

class trieNode(treeNode):
    
    ''' A subclass of treeNode that allows for checking 
    non-zero members amongst children branches and other
    conveniences. '''
    
    def getParentNode(self): return self.parent.parent
    
    def setChildNode(self,child,newchild):
        
        for branch in self.children:
            if branch.child != None and branch.child == child:
                branch.child = newchild
                if newchild: newchild.parent = branch
                return
    
    def iterNonEmptyChildrenNodes(self):    
        
        ''' Iterate over all children that are not empty. '''
        
        for branch in self.children:
            if (branch.child != None):
                yield branch.child
                
    def getNonEmptyChildrenNodes(self):
        
        ''' Acquire a list of all non-empty children. '''
        
        return [x for x in self.iterNonEmptyChildrenNodes()]
    
    def getNonEmptyChildrenBranches(self):
        
        ''' Acquire a list of all non-empty children. '''
    
        return [x.parent for x in self.iterNonEmptyChildrenNodes()]
    
    def getNonEmptyChildrenBranchLabels(self):
        
        return [x.parent.label for x in self.iterNonEmptyChildrenNodes()]
    
    def numEmptyChildrenNodes(self):
        
        ''' Acquire the number of children nodes that are marked 0 
        or nonexistent. '''
        
        empty = 0
        
        for branch in self.children:
            if (branch.child == None or branch.child.label == 0):
                empty += 1
                
        return empty

class trie(Sized,treeStructure):
    
    ''' Defines a trie across a range of strings. '''
    
    alphabet  = None
    root      = None
    count     = 0
    nextLabel = 1
    
    def __init__(self):
        
        self.alphabet = []
        self.root = trieNode(lbl=0)
        
    def __contains__(self,x):

        ''' Implementing for interface (Container). '''
        
        return self.search(x)
        
    def __len__(self):
        
        ''' Implementing for interface (Sized). '''
        
        return self.count
        
    def getAlphabet(self): return self.alphabet
    def getRoot(self): return self.root
    
    def _makeNewChildForNodes(self,node,char):
        
        ''' Called when a new item is added to the alphabet. Adds a new
        empty node to all children. '''
        
        if node == None: return
        if len(node.getChildren()) != 0:
            for child in node.getChildren():
                self._makeNewChildForNodes(child.child,char)
        node.addChild(treeBranch(parent=node,label=char))
    
    def _newNode(self,parent):
        
        ''' Create a new node. '''
        
        t = trieNode(parent=parent,lbl=0)
        for char in self.alphabet:
            t.addChild(treeBranch(parent=t,label=char))
        return t
    
    def _deleteNode(self,node,child=None):
        
        ''' Delete a node. '''
        
        # If a child was given, clear it.
        if child != None:
            node.setChildNode(child,None)
            child.parent = None
            
        # If this is root, stop.
        if node == self.root: return
        
        # Check node for emptiness.
        numempty = node.numEmptyChildrenNodes()
        if (numempty == len(self.alphabet)):
            # All empty.
            self._deleteNode(node.getParentNode(),node)
    
    def search(self,seq):
        
        ''' Search for a sequence in the trie. Returns true
        if it exists. '''
        
        index  = 0
        cursor = self.getRoot()
        
        for character in seq:
            
            # In alphabet?
            if not character in self.alphabet:
                return False
            
            # Get index in alphabet.
            index = self.alphabet.index(character)
            
            # Traverse trie.
            nextitem = cursor.getChildByIndex(index)
            if (nextitem.child == None):
                return False
            cursor = nextitem.child
        
        return (cursor.label > 0)
    
    def insert(self,seq):
        
        ''' Dynamically insert a sequence into the trie. '''
    
        index  = 0
        cursor = self.getRoot()
        self.count += 1
        
        for character in seq:
            
            # In alphabet?
            if not character in self.alphabet:
                self.alphabet.append(character)
                self._makeNewChildForNodes(self.getRoot(),character)
            
            # Get index in alphabet.
            index = self.alphabet.index(character)
            
            # Traverse trie.
            nextitem = cursor.getChildByIndex(index)
            if (nextitem.child == None):
                nextitem.child = self._newNode(nextitem)
            cursor = nextitem.child
        
        # Mark the leaf.
        cursor.label = self.nextLabel
        self.nextLabel += 1
        
        # Return its label.
        return cursor.label
        
    def delete(self,seq):
        
        ''' Remove a sequence from the trie. Will not remove
        added characters to alphabet. '''
        
        index  = 0
        cursor = self.getRoot()
        
        for character in seq:
            
            if not character in self.alphabet:
                raise LookupError('Sequence %s not in trie.' % (seq))
            index = self.alphabet.index(character)
            
            # Traverse trie.
            nextitem = cursor.getChildByIndex(index)
            if (nextitem.child == None):
                raise LookupError('Sequence %s not in trie.' % (seq))
            cursor = nextitem.child
        
        # Work way up and check for completely empty subtrees.
        self.count -= 1
        self._deleteNode(cursor)
        
    def _createStringLinesForNode(self,root,levelsDeep=-1):
        
        ''' Private method for stringifying this object. '''
        
        if (root == None): return ''
        elif (root != self.root):
            strOut   = '  '*levelsDeep + '| ' + root.parent.label + '\n'
        else: strOut = ''
        for branch in root.getChildren():
            strOut += self._createStringLinesForNode(branch.child,levelsDeep+1)
        return strOut     

class patriciaTree(trie):
    
    ''' Defines a PATRICIA tree (condensed trie) across a range of strings. '''
    
    def __contains__(self,x): return (self.search(x) != 0)
    
    def _deleteNode(self,node,child=None):
        
        ''' Merges node upward if it only has 1 child. '''
        
        # If a child was given, clear it.
        if child is not None:
            node.setChildNode(child,None)
            child.parent = None
            
        # If this is root, stop.
        if node is self.root: return
        
        # Check node for emptiness.
        numEmpty = node.numEmptyChildrenNodes()
        alphaLen = len(self.alphabet)
        if (numEmpty == alphaLen - 1):
            # Only has a single child. Merge upwards.
            parent     = node.getParentNode()
            edge       = node.parent
            nonemptybr = node.getNonEmptyChildrenBranches()[0]
            edge.label = edge.label + nonemptybr.label
            parent.setChildNode(node,nonemptybr.child)    
        elif (numEmpty == alphaLen):        
            # All empty.
            self._deleteNode(node.getParentNode(),node)
    
    def _query(self,seq):
        
        ''' Acquires the leaf for a sequence if it exists. Returns the
        node and its parent of wherever failure or success occurs. Returns
        a tuple of elements where cursor is the node the query stopped at,
        parentofcursor is the branch associated with that node, cpos is the
        position of seq that the query stopped at, and spos is where cpos was
        last at before finding a mismatch. '''
        
        index,cpos,spos = 0,0,0
        cursor = self.getRoot()
        parentofcursor = cursor
        seqLength = len(seq)
        
        while (cpos < seqLength):

            # Get character.
            character = seq[cpos]
            if not character in self.alphabet:
                return None, None, cpos, spos
            index = self.alphabet.index(character)
            
            # Go to next respective item in trie.
            nextitem       = cursor.getChildByIndex(index)
            parentofcursor = nextitem
            if (nextitem.child == None):
                # Item is not in the trie.
                return None, parentofcursor, cpos, spos
            cursor = nextitem.child 
            label  = nextitem.label
            spos   = cpos
            lpos   = 0
            
            # Check character matching.
            labelLength = len(label)
            while (lpos < labelLength):
                if (cpos >= seqLength):
                    # Still more in the label! Not found.
                    return None, parentofcursor, cpos, spos
                elif (label[lpos] != seq[cpos]):
                    # Found a mismatch.
                    return None, parentofcursor, cpos, spos
                lpos += 1
                cpos += 1
                
        return cursor, parentofcursor, seqLength, spos
    
    def search(self,seq):
        
        ''' Search for a sequence in the PATRICIA tree. Returns its position in
        addition sequence if it exists. Else, returns 0. '''
        
        query,_,_,_ = self._query(seq)
        if (query is None): return 0
        return query.label       
    
    def insert(self,seq):
        
        ''' Dynamically insert a sequence into the PATRICIA tree. Returns the
        unique index in the tree for that string. '''
        
        # Check alphabet membership.
        for character in seq:
            if not character in self.alphabet:
                self.alphabet.append(character)
                self._makeNewChildForNodes(self.getRoot(),character)        
        
        # Perform a query.
        query,queryparent,cpos,spos = self._query(seq)
        if query == None or query.label == 0: self.count += 1
        else: return 0 # Already in the tree.
        
        prevLabel = ''
        # Insert Case 1: Middle of Edge. Check for label of edge.
        currEdge = queryparent
        cmpLabel = seq[spos:cpos+1]
        
        if (len(currEdge.label) > 1 and cmpLabel != currEdge.label):
            
            # Know point of convergence.
            newLabel = seq[spos:cpos]
            oldLabel = currEdge.label
            
            # Replace previous edge label.
            currEdge.label = newLabel
            
            # See what has to be split to a new edge.
            extent = oldLabel[len(newLabel):]
            
            # Make new node to join split edge.
            oldNode = currEdge.child
            newNode = self._newNode(currEdge)
            currEdge.child = newNode
            
            # Get ready for insertion to newChild.
            extentIndex  = self.alphabet.index(extent[0])
            extentBranch = newNode.getChildByIndex(extentIndex)            
            
            # Special Case: Adding a subset of an existing entry.
            if cpos >= len(seq):
                # Simply mark the new node.
                newNode.label = self.nextLabel
                insertBranch  = None
            
            # Add extent to newChild as well as this insertion.
            else:
                insertIndex  = self.alphabet.index(seq[cpos])
                insertBranch = newNode.getChildByIndex(insertIndex)
            extentBranch.label = extent
            extentBranch.child = oldNode
            oldNode.parent = extentBranch            
            
        # Insert Case 2: New Node.
        else: insertBranch = queryparent
        
        if insertBranch != None:
            prevLabel = insertBranch.label
            if cpos != len(seq):
                insertBranch.label = seq[cpos:]
            if insertBranch.child is None:
                insertBranch.child = self._newNode(insertBranch)
            insertBranch.child.label = self.nextLabel
        
        self.nextLabel += 1
        return self.nextLabel-1

    def delete(self,seq):
        
        ''' Remove a sequence from the PATRICIA tree. Will not remove 
        added characters to alphabet. '''

        query,_,_,_ = self._query(seq)
        
        if query is None: raise LookupError('Sequence %s not in trie.' % (seq))
        
        # Work way up and check for completely empty subtrees.
        self.count -= 1
        self._deleteNode(query)



''' Defines an interface to manage interfacing with the system for respective
application calls and implements some of these for executables such as FastTree
and RAxML. Currently requires a UNIX-like environment (e.g., Mac OS X or a
Linux-based environment). '''

# Date:   Oct 16 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

import tree
from base import longest_common_substring
from abc import ABCMeta as abstractclass, abstractmethod
from os import mkdir, getcwd, chdir, name as os_name
from os.path import abspath, isdir, isfile
from subprocess import call, PIPE, Popen as system
from tempfile import NamedTemporaryFile

# Constants

E_FASTTREE = 'fasttree'
E_RAXML    = 'raxmlHPC'
E_TREEPUZZ = 'puzzle'
E_RSPR     = 'rspr'
L_TEMPDIR  = 'pylogeny'

# Executable Existence Function

def exeExists(cmd):
    
    ''' Determines whether a function exists in a UNIX environment. '''
  
    return call('type %s' % (cmd),shell=True,stdout=PIPE,stderr=PIPE) == 0

# Temporary Directory Context

class aTemporaryDirectory(object):
    
    ''' A class intended to be used as a context manager that allows
    Python to run in a temporary directory for a finite period of time. '''
    
    def __init__(self,dir=None):
        
        self.path = '/tmp/' + dir + '/'
        if not isdir(self.path): mkdir(self.path)
        self.current = getcwd()
        
    def __enter__(self):      chdir(self.path)
    def __exit__(self,*args): chdir(self.current)

# Executable Interface

class executable(object):
    
    ''' An abstract class for the instantation and running of a single instance
    for a given application. '''

    exeName = None
    __metaclass__ = abstractclass
    
    @abstractmethod
    def getInstructionString(self): pass
    
    def run(self):
        
        ''' Perform a run of this application. '''
        
        if (os_name != 'posix'):
            import platform
            pf = platform.system()
            raise SystemError('No UNIX-like environment. ' + 
                              "You are currently using platform '%s', " % (pf)+
                              'and cannot run a binary in a shell.')
        elif (self.exeName and not exeExists(self.exeName)):
            raise SystemError("'%s' is not installed on your system." % (
                self.exeName))
        sproc = system(self.getInstructionString(),stdout=PIPE,
                       stderr=PIPE,shell=True)
        return sproc.communicate()[0]

# Executable Classes

class treepuzzle(executable):
    
    ''' Wrap TREE-PUZZLE in order to create an intermediate file for CONSEL to
    read and assign confidence to a set of trees. Requires TREE-PUZZLE to be
    installed. '''

    exeName = E_TREEPUZZ

    def __init__(self,ali,treefile):
        self.treefile  = treefile
        self.alignment = ali
        if self.alignment == None: raise AttributeError('No alignment defined.')
        elif self.treefile == None: raise AttributeError('No treefile defined.')

    def getInstructionString(self):
        return 'echo "y" | %s %s %s -wsl' % (
            self.exeName,self.alignment.getPhylip(),self.treefile)
    
    def getSiteLikelihoodFile(self):
        self._output = self.run()
        if not isfile('%s.sitelh' % (self.treefile)):
            raise IOError("No site-likelihood output. Executed '%s' (%s)." % (
                self.getInstructionString(),self._output))
        else: return '%s.sitelh' % (self.treefile) 

class consel(executable):
    
    ''' Denotes a single run of the CONSEL workflow in order to acquire a
    confidence interval and perform an AU test on a set of trees. Requires
    CONSEL to be installed. '''
    
    def __init__(self,treeset,alignment,name):
        
        self.treeset   = treeset
        self.name      = name
        self.alignment = alignment
        self.sitelh    = None
        self.raw       = None
        self.rmt       = None
        self.pv        = None
        self.auvals    = list()
        self.interval  = tree.treeSet()
        self.rejected  = tree.treeSet()
        self.instruction = ''
        
    def getInstructionString(self): return self.instruction
    
    def _convertRawData(self):
        
        if not isfile('%s.mt' % (self.name)):
            self.instruction = 'seqmt --puzzle %s %s.mt' %\
                (self.sitelh,self.name)
            self._out = self.run()
            if not isfile('%s.mt' % (self.name)):
                raise IOError("CONSEL 'seqmt' did not create mt file.") 
        self.raw = '%s.mt' % (self.name)
        
    def _createReplicates(self):
        
        if not isfile('%s.rmt' % (self.name)):
            self.instruction = 'makermt %s' % (self.raw)
            self._out = self.run()
            if not isfile('%s.rmt' % (self.name)):
                raise IOError("CONSEL 'markermt' did not create rmt file.")
        self.rmt = '%s.rmt' % (self.name)
        
    def _run(self):
        
        if not isfile('%s.pv' % (self.name)):
            self.instruction = 'consel %s.rmt' % (self.name)
            self._out = self.run()
        self.pv = '%s.pv' % (self.name)
        
    def _getAU(self):
        
        self.instruction = 'catpv %s' % (self.pv)
        pvout = self.run()
        for line in pvout.split('\n'):
            spl = line.split()
            if len(spl) < 3: continue
            elif not spl[2].isdigit(): continue
            it = (int(spl[2])-1,float(spl[4]))
            self.auvals.append(it)
            if it[1] >= 0.05: self.interval.addTree(self.treeset[it[0]])
            else: self.rejected.addTree(self.treeset[it[0]])    
        self._out = pvout
    
    def getInterval(self):
        
        ''' Compute the AU test. Return the interval of trees. '''
        
        with aTemporaryDirectory(L_TEMPDIR):
            if not isfile('%s.trees.sitelh' % (self.name)):
                self.treefile = self.treeset.toTreeFile(self.name + '.trees')
                self.treepuzz = treepuzzle(self.alignment,self.treefile)
                self.sitelh   = self.treepuzz.getSiteLikelihoodFile()
            else: self.sitelh = '%s.trees.sitelh' % (self.name)
            self._convertRawData()
            self._createReplicates()
            self._run()
            self._getAU()
        return self.interval

class fasttree(executable):
    
    ''' Denotes a single run of the FastTree executable in order to acquire an
    approximate maximum likelihood tree for the input alignment. See
    http://www.microbesonline.org/fasttree/ for more information on FastTree.
    Requires FastTree to be installed. '''
    
    exeName = E_FASTTREE
    
    def __init__(self,inp_align,out_file=None,isProtein=True):
        self.alignment = inp_align
        self.isProtein = isProtein
        self.out       = out_file
        if (self.alignment == None):
            raise IOError('No alignment defined.')
        
    def getInstructionString(self):
        s = '%s -quiet -nopr < %s' % (self.exeName,self.alignment)
        if self.out: s += ' > %s' % (self.out)
        return s
    
class raxml(executable):
    
    ''' Denotes a single run of the RAxML executable. See
    http://sco.h-its.org/exelixis/software.html for more information on RAxML.
    Requires RAxML to be installed. '''
    
    exeName = E_RAXML
    
    def __init__(self,inp_align,out_file,model=None,
                 is_Protein=True,interTrees=False,
                 alg=None,startingTree=None,rapid=False,
                 slow=False,optimizeBootstrap=False,numboot=100,
                 log=None,wdir=None):
        self.alignment     = inp_align
        self.out           = out_file
        self.model         = model
        self.protein       = is_Protein
        self.intermediates = interTrees
        self.alg           = alg
        self.startingTree  = startingTree
        self.slowBoot      = slow
        self.rapidBoot     = rapid
        self.numBoot       = numboot
        self.optBoot       = optimizeBootstrap
        self.log           = log
        self.workdir       = abspath(wdir)
        if not self.model:
            if is_Protein: self.model = 'PROTGAMMAJTT'
            else:          self.model = 'GTRGAMMA'
            
    def getInstructionString(self):
        s = '%s -s %s -n %s -m %s' %\
            (self.exeName,self.alignment,self.out,self.model)
        if self.alg: s += ' -f %s' % (self.alg)
        if self.intermediates: s += ' -j'
        if self.startingTree: s += ' -t %s' % (self.startingTree)
        if self.rapidBoot or self.slowBoot:
            if self.rapidBoot:  s += ' -x 234534251 -N %d' % (self.numBoot)
            elif self.slowBoot: s += ' -b 234534251 -N %d' % (self.numBoot)
        if self.optBoot: s += ' -k'
        if self.workdir: s += ' -w %s' % (self.workdir)
        if self.log: s += ' > %s' % (self.log)
        return s
    
    def runFunction(self,alg):
        self.alg = alg
        self.run()

class rspr(executable):

    ''' Denotes a single run of the rSPR executable by Dr. Chris Whidden (2014),
    a software package for computing rooted subtree-prune-and-regraft (SPR)
    distances. See http://kiwi.cs.dal.ca/Software/RSPR. Requires the executable
    to be on PATH. '''

    exeName = E_RSPR
    RSPR_ALG_DEFAULT = ''
    RSPR_ALG_APPROX  = '-approx'
    RSPR_ALG_BB      = '-bb'
    RSPR_ALG_FPT     = '-fpt'

    def __init__(self,treeA,treeB,algorithm=RSPR_ALG_DEFAULT,overlap=True):
        
        ''' Algorithm choices are defined in this class. If overlap is set to
        True, will attempt to consolidate taxa names such that they are
        overlapping (otherwise, RSPR will return an error if they do not match).
        '''
        
        self.inputA   = treeA
        self.inputB   = treeB
        self.treeFile = ''
        self.algthm   = algorithm
        self._output  = ''
        if overlap: self._overlapTaxa(treeA,treeB)
        self._makeTemporaryFile()
        
    def _makeTemporaryFile(self):
        t1,t2 = self.inputA,self.inputB
        fi    = NamedTemporaryFile(delete=False)
        fi.close()
        o = open(fi.name,'w')
        o.write('%s\n%s' % (t1.newick,t2.newick))
        o.close()
        self.treeFile = fi.name        
        
    def _overlapTaxa(self,a,b):
        topol = a.toTopology()
        other = b.toTopology()
        anodes = topol.getAllLeaves()
        bnodes = other.getAllLeaves()
        if (len(anodes) != len(bnodes)):
            raise IOError('Non-overlapping taxa lengths.')
        runnli = []
        for n in anodes:
            corresp    = None
            correspstr = ''
            for i in [x for x in bnodes if not x in runnli]:
                lcstr = longest_common_substring(n.label,i.label)
                if (len(lcstr) > len(n.label)*(3./4) 
                    and len(lcstr) > len(i.label)*(3./4) and
                    len(lcstr) > len(correspstr)):
                    corresp    = i
                    correspstr = lcstr
            if corresp:
                n.label       = correspstr
                corresp.label = correspstr
                runnli.append(corresp)
            else:
                raise IOError('Could not find overlapping taxa for %s.' 
                              % (n.label))
        self.inputA = topol.toTree()
        self.inputB = other.toTree()         

    def getSPRDistance(self):
        self._output = self.run()
        return (int(self._output.split('drSPR=')[-1]))

    def getInstructionString(self):
        return '%s %s < %s' % (self.exeName,self.algthm,self.treeFile)
